use std::{collections::BTreeMap, hash::Hasher};

#[derive(Debug, Clone, Copy)]
pub struct Index {
    pub start: u32,
    pub end: u32,
}

impl Index {
    pub fn new(start: u32, end: u32) -> Self {
        Self { start, end }
    }
}

#[derive(Debug)]
pub struct StringTable {
    bytes: Vec<u8>,
    indices: BTreeMap<u64, Index>,
}

impl core::ops::Index<Index> for StringTable {
    type Output = [u8];

    fn index(&self, idx: Index) -> &Self::Output {
        &self.bytes[idx.start as usize..idx.end as usize]
    }
}

impl StringTable {
    pub fn new() -> Self {
        Self {
            bytes: Vec::new(),
            indices: BTreeMap::new(),
        }
    }

    pub fn get_str(&self, idx: Index) -> &str {
        unsafe { core::str::from_utf8_unchecked(&self[idx]) }
    }

    pub fn get_bytes(&self, idx: Index) -> &[u8] {
        &self[idx]
    }

    pub fn insert<B: AsRef<[u8]>>(&mut self, bytes: B) -> Index {
        let bytes = bytes.as_ref();
        let hash = {
            let mut hasher = std::hash::DefaultHasher::new();
            hasher.write(bytes);
            hasher.finish()
        };

        if let Some(idx) = self.indices.get(&hash).cloned() {
            idx
        } else {
            self.insert_inner(hash, bytes)
        }
    }

    fn insert_inner(&mut self, hash: u64, bytes: &[u8]) -> Index {
        let start = self.bytes.len();
        self.bytes.extend(bytes);
        let end = self.bytes.len();
        let index = Index::new(start as u32, end as u32);
        self.indices.insert(hash, index);
        index
    }
}
